#ifdef RICH_MPI
#include "source/mpi/MeshPointsMPI.hpp"
#endif
#include <iostream>
#include <fstream>
#include <cmath>
#include <vector>
#include "source/misc/int2str.hpp"
#include "source/tessellation/VoronoiMesh.hpp"
#include "source/tessellation/RoundGrid.hpp"
#include "source/tessellation/geometry.hpp"
#include "source/newtonian/two_dimensional/hdsim2d.hpp"
#include "source/tessellation/tessellation.hpp"
#include "source/newtonian/common/hllc.hpp"
#include "source/newtonian/common/ideal_gas.hpp"
#include "source/tessellation/VoronoiMesh.hpp"
#include "source/newtonian/two_dimensional/spatial_distributions/uniform2d.hpp"
#include "source/newtonian/two_dimensional/point_motions/round_cells.hpp"
#include "source/newtonian/two_dimensional/point_motions/lagrangian.hpp"
#include "source/newtonian/two_dimensional/point_motions/eulerian.hpp"
#include "source/newtonian/two_dimensional/point_motions/CustomMotion.hpp"
#include "source/misc/utils.hpp"
#include "source/newtonian/two_dimensional/source_terms/zero_force.hpp"
#include "source/newtonian/two_dimensional/geometric_outer_boundaries/SquareBox.hpp"
#include "source/newtonian/two_dimensional/diagnostics.hpp"
#include "source/misc/simple_io.hpp"
#include "source/newtonian/two_dimensional/hdf5_diagnostics.hpp"
#include "source/newtonian/two_dimensional/condition_action_sequence.hpp"
#include "source/newtonian/two_dimensional/simple_extensive_updater.hpp"
#include "source/newtonian/two_dimensional/simple_cell_updater.hpp"
#include "source/misc/vector_initialiser.hpp"
#include "source/newtonian/two_dimensional/simple_flux_calculator.hpp"
#include "source/misc/mesh_generator.hpp"
#include "source/newtonian/two_dimensional/amr.hpp"
#include "source/newtonian/two_dimensional/stationary_box.hpp"
#include "source/newtonian/test_2d/multiple_diagnostics.hpp"
#include "source/newtonian/test_2d/consecutive_snapshots.hpp"
#include "source/newtonian/test_2d/main_loop_2d.hpp"

using namespace std;
using namespace simulation2d;

namespace
{

  double calc_ambient_density(double y)
  {
    return pow(fmax(-y,1e-4),1.5);
  }

#ifdef RICH_MPI
	vector<Vector2D> process_positions(const SquareBox& boundary)
	{
		int ws=0;
		Vector2D lower_left=boundary.getBoundary().first;
		Vector2D upper_right=boundary.getBoundary().second;
		MPI_Comm_size(MPI_COMM_WORLD,&ws);
		vector<Vector2D> res = RandSquare(ws,
				lower_left.x, upper_right.x,
				lower_left.y, upper_right.y);
		return res;
	}
#endif // RICH_MPI

	vector<ComputationalCell> calc_init_cond
		(const Tessellation& tess)
	{
		vector<ComputationalCell> res(static_cast<size_t>(tess.GetPointNo()));
		for (size_t i = 0; i < res.size(); ++i) {
		  const Vector2D r = tess.GetCellCM(static_cast<int>(i));
		  const bool is_wedge = pow(r.y+0.5,2)+pow(r.x,2)<0.01;
		  res[i].density = calc_ambient_density(r.y);
		  res[i].pressure = 1e-9;
		  res[i].velocity = is_wedge ? Vector2D(0,0) : Vector2D(1,0);
		  res[i].stickers.push_back(is_wedge);
		}
		write_number(atan(0.1), "wedge_angle.txt");
		return res;
	}

	class LeftBoundaryEdge : public ConditionActionSequence::Condition
	{
	public:

		LeftBoundaryEdge(void) {}

		pair<bool, bool> operator()
			(const Edge& edge,
				const Tessellation& tess,
				const vector<ComputationalCell>& /*cells*/,TracerStickerNames const& /*ts*/) const
		{
			if (tess.GetOriginalIndex(edge.neighbors.first) == tess.GetOriginalIndex(edge.neighbors.second))
			{
				if (abs(edge.vertices.first.x-edge.vertices.second.x)<abs(edge.vertices.first.y-edge.vertices.second.y))
				{
					if(edge.neighbors.second<tess.GetPointNo())
					{
						Vector2D const& r = tess.GetMeshPoint(edge.neighbors.second);
						if(r.x>edge.vertices.first.x)
							return pair<bool, bool>(true, false);
					}
					else
					{
						Vector2D const& r = tess.GetMeshPoint(edge.neighbors.first);
						if(r.x>edge.vertices.first.x)
							return pair<bool, bool>(true, true);
					}
				}
				return pair<bool, bool>(false, false);
			}
			return pair<bool, bool>(false, false);
		}
	};

	void conserved_to_extensive
		(const Conserved& c, const ComputationalCell& /*cell*/, Extensive &res)
	{
		res.mass = c.Mass;
		res.momentum = c.Momentum;
		res.energy = c.Energy;
	}

	class ConstantGhost : public ConditionActionSequence::Action
	{
	public:

		ConstantGhost(const RiemannSolver& rs) :
			rs_(rs) {}

		void operator()
			(const Edge& edge,
				const Tessellation& tess,
				const Vector2D& /*edge_velocity*/,
				const vector<ComputationalCell>& cells,
				const EquationOfState& eos,
				const bool aux,
				Extensive &res,double /*time*/,TracerStickerNames const& ts) const
		{

		  const Vector2D centroid = 0.5*(edge.vertices.first + edge.vertices.second);
		  ComputationalCell ghost;
		  ghost.density = calc_ambient_density(centroid.y);
		  ghost.pressure = 1e-9;
		  ghost.velocity = Vector2D(1,0);

			if (aux)
				assert(edge.neighbors.first < tess.GetPointNo());
			else
				assert(edge.neighbors.second < tess.GetPointNo());
			const Vector2D p = normalize
				(edge.vertices.second -
					edge.vertices.first);
			const Vector2D n = normalize
				(remove_parallel_component
					(aux ?
						edge.vertices.first - tess.GetMeshPoint(edge.neighbors.first) :
						tess.GetMeshPoint(edge.neighbors.second) - edge.vertices.first,
						p));
			const double v = 0;
			const pair<ComputationalCell, ComputationalCell> cc_left_righ =
				aux ?
				pair<ComputationalCell, ComputationalCell>
				(cells.at(static_cast<size_t>(edge.neighbors.first)), ghost) :
				pair<ComputationalCell, ComputationalCell>
				(ghost, cells.at(static_cast<size_t>(edge.neighbors.second)));
			const pair<Primitive, Primitive> left_right =
				pair<Primitive, Primitive>
			  (convert_to_primitive(cc_left_righ.first, eos, ts),
			   convert_to_primitive(cc_left_righ.second, eos, ts));
			const Conserved c = rotate_solve_rotate_back
				(rs_,
					left_right.first,
					left_right.second,
					v, n, p);
			conserved_to_extensive(c, ghost, res);
		}

	private:
		const RiemannSolver& rs_;
	};

	class SimData
	{
	public:

		SimData(double adiabatic_index = 5. / 3.,
			double width = 2) :
			pg_(),
			outer_(-width / 2, width / 2, width / 2, -width / 2),
			tess_(RoundGrid(RandSquare(75*75,-width / 2, width / 2, -width / 2, width / 2),&outer_),
			      outer_),
			eos_(adiabatic_index),
			rs_(),
			point_motion_(),
			evc_(),
			force_(),
			tsf_(0.3),
			fc_(VectorInitialiser<pair<const ConditionActionSequence::Condition*, const ConditionActionSequence::Action*> >
				(pair<const ConditionActionSequence::Condition*, const ConditionActionSequence::Action*>
					(new LeftBoundaryEdge, new ConstantGhost(rs_)))
				(pair<const ConditionActionSequence::Condition*, const ConditionActionSequence::Action*>
					(new IsBoundaryEdge, new FreeFlowFlux(rs_)))
				(pair<const ConditionActionSequence::Condition*, const ConditionActionSequence::Action*>
					(new RegularSpecialEdge("wedge"), new RigidWallFlux(rs_)))
				(pair<const ConditionActionSequence::Condition*, const ConditionActionSequence::Action*>
					(new IsBulkEdge, new RegularFlux(rs_)))()),
			eu_(),
			cu_(VectorInitialiser<pair<const SimpleCellUpdater::Condition*, const SimpleCellUpdater::Action*> >
				(pair<const SimpleCellUpdater::Condition*, const SimpleCellUpdater::Action*>
					(new HasSticker("wedge"), new SkipUpdate))()),
			sim_
			(
#ifdef RICH_MPI
				proctess_,
#endif // RICH_MPI
				tess_,
				outer_,
				pg_,
				calc_init_cond(tess_),
				eos_,
				point_motion_,
				evc_,
				force_,
				tsf_,
				fc_,
				eu_,
				cu_,
				TracerStickerNames (vector<string>(),vector<string>(1,"wedge")))
		{
			write_number(adiabatic_index,
				"adiabatic_index.txt");
			write_number(4.0 / sqrt(5. / 3.),
				"mach_number.txt");
		}

		hdsim& getSim(void)
		{
			return sim_;
		}

	private:
		const SlabSymmetry pg_;
		SquareBox outer_;
#ifdef RICH_MPI
		VoronoiMesh proctess_;
#endif // RICH_MPI
		VoronoiMesh tess_;
		const IdealGas eos_;
		const Hllc rs_;
	  Eulerian point_motion_;
		const StationaryBox evc_;
		ZeroForce force_;
		const SimpleCFL tsf_;
		const ConditionActionSequence fc_;
		const SimpleExtensiveUpdater eu_;
		const SimpleCellUpdater cu_;
		hdsim sim_;
	};

  class WriteCycle: public DiagnosticFunction
  {
  public:

    WriteCycle(const string& fname):
      fname_(fname) {}

    void operator()(const hdsim& sim)
    {
      write_number(sim.getCycle(),fname_);
    }

  private:
    const string fname_;
  };

	void my_main_loop(hdsim& sim)
	{
	  const int max_iter = 5e6;
	  const double tf = 1e1;
	  SafeTimeTermination term_cond(tf,max_iter);

	  MultipleDiagnostics diag
	    (VectorInitialiser<DiagnosticFunction*>()
	     (new ConsecutiveSnapshots(new ConstantTimeInterval(tf/100),
				       new Rubric("output/snapshot_",".h5")))
	     (new WriteTime("time.txt"))
	     (new WriteCycle("cycle.txt"))());

	  main_loop(sim,
		    term_cond,
		    &hdsim::TimeAdvance,
		    &diag);
	}
}

int main(void)
{
#ifdef RICH_MPI
	MPI_Init(NULL,NULL);
#endif // RICH_MPI

	SimData sim_data;

	hdsim& sim = sim_data.getSim();

	write_snapshot_to_hdf5(sim, "output/initial.h5");

	my_main_loop(sim);

	write_snapshot_to_hdf5(sim, "output/final.h5");

	return 0;
}


